/*
 * Copyright (C) 2025 sillysagiri
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
 */
 
#pragma once

#include "stb_image.h"
#include <cstdint>
#include <filesystem>
#include <string>
#include <vector>

struct Metadata {
	uint64_t header = 0x676D69796C6C6973;
	int8_t version = 14;
	int8_t format;
	int16_t width;
	int16_t height;
	int16_t palette_count;
	int32_t original_size;
	int32_t compress_size;
};

enum Format
{
	Format_RGB_256 = 0,
	Format_RGB_16,
	Format_INDEXED_4,
	Format_INDEXED_16,
	Format_INDEXED_256,
	Format_INDEXED_32A8,
	Format_INDEXED_8A32,
	Format_PALETTE_16,
};

static inline uint16_t RGB255_RGB16(uint8_t r, uint8_t g, uint8_t b, uint8_t a = 255, uint8_t alpha_threshold = 128) {
	const uint16_t alpha_bit = (a >= alpha_threshold) ? 0b1000000000000000 : 0;
	// const uint16_t alpha_bit = 0b1000000000000000;
	return alpha_bit
			| (((uint8_t)r >> 3) & 0b00011111)
			| (((uint8_t)g >> 3) & 0b00011111) << 5
			| (((uint8_t)b >> 3) & 0b00011111) << 10;
}

struct Image {
	int width = 0, height = 0;
	std::vector<uint8_t> data;

	~Image() {}
	void Load(const std::filesystem::path &path);
};

struct Color {
	uint8_t r, g, b, a;
};

typedef std::vector<uint8_t> ImageMapped;
typedef std::vector<Color> Palette;

struct Options {
	std::string path_input;
	std::string path_output;
	std::string format;
	std::string palette_format;
	// std::string compression;
	bool enable_be = false;
};

void WriteOutput(const Options &opt, const Metadata &meta, const Palette &palette, void *image_buffer);
ImageMapped Quantize(const Image &image, Palette &palette_output, int num_colors, bool dither, uint8_t format);

static inline uint16_t to_be_int16(int16_t value) {
    uint16_t uval = static_cast<uint16_t>(value);
    return (uval >> 8) | (uval << 8);
}

static inline uint32_t to_be_int32(int32_t value) {
    uint32_t uval = static_cast<uint32_t>(value);
    return ((uval >> 24) & 0x000000FF) |
           ((uval >> 8)  & 0x0000FF00) |
           ((uval << 8)  & 0x00FF0000) |
           ((uval << 24) & 0xFF000000);
}

static inline uint16_t to_be_int16(uint16_t value) {
    return (value >> 8) | (value << 8);
}

static inline uint32_t to_be_int32(uint32_t value) {
    return ((value >> 24) & 0x000000FF) |
           ((value >> 8)  & 0x0000FF00) |
           ((value << 8)  & 0x00FF0000) |
           ((value << 24) & 0xFF000000);
}